#!/usr/bin/perl
# TCP via webproxy
#	listen on a local port
#	on accept
#		open to proxyhost:port
#		do necessary protocol startup to get remotehost:port
#		read/write loop
use IO::Socket;
use IO::Select;

$prog=$0;
$prog =~ s,.*/,,;
$| = 1;

$rdbufsize=65536;
$auth="";
$lflg=0;
$nflg=0;
$vflg=0;
while ($ARGV[0] =~ /^-./) {
	if ($ARGV[0] eq "-v") {
		$vflg++;
	}
	if ($ARGV[0] eq "-n") {
		$nflg++;
	}
	if ($ARGV[0] eq "-l") {
		$lflg++;
	}
	if ($ARGV[0] eq "-a") {
		$auth=$ARGV[1];
		shift;
	}
	shift;
}

# review args to determine process as sender or listener
$listenport=$ARGV[0];
if ($listenport !~ /^[0-9]+$/) {
	$err=1;
} else {
	$proxyhp=$ARGV[1];
	if ($proxyhp !~ /.:[0-9]+$/) {
		$err=1;
	} else {
		$remotehp=$ARGV[2];
		if ($remotehp !~ /.:[0-9]+$/) {
			$err=1;
		} else {
			$err=0;
		}
	}
}
if ($err) {
	print "Usage: $prog [-a <auth>] [-l] [-n] [-v] listenport proxyhost:port remotehost:port\n";
	print "\t-a\tuuencoded authentication of user:pass\n";
	print "\t-l\tloop listener - allow multiple sessions\n";
	print "\t-v\tverbose, repeat (eg -v -v) for more verbose\n";
	print "\t-n\tdo not do DNS lookups (in verbose mode)\n";
	exit 1
}
($proxyhost, $proxyport)=split(":", $proxyhp);

unless ($fdsock = IO::Socket::INET->new(Listen => 5, LocalPort => $listenport, Proto => 'tcp', ReuseAddr => 1)) {
	printf "$prog: Cannot listen on port %d: %s", $listenport, $!."\n";
	exit 1;
}

# main loop
$chanmax=0;
while (1) {
	# create select list
	$sel = IO::Select->new();
	# add listener
	$cnt=0;
	if ($fdsock != -1) {
		if ($vflg > 1) {
			@lt=localtime(time());
			printf "%04d-%02d-%02d %02d:%02d:%02d select: Add listener\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0];
		}
		$sel->add($fdsock);
		$cnt++;
	}
	# add open file descriptors
	for ($chan=0; $chan < $chanmax; $chan++) {
		if ($fdir[$chan] != -1) {
			if ($chstate[$chan] == 4) {
				if ($vflg > 1) {
					@lt=localtime(time());
					printf "%04d-%02d-%02d %02d:%02d:%02d select: Add channel %d input side\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan;
				}
				$sel->add($fdir[$chan]);
				$cnt++;
			} else {
				if ($vflg > 1) {
					@lt=localtime(time());
					printf "%04d-%02d-%02d %02d:%02d:%02d select: Not adding channel %d input side (state=%d)\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $chstate[$chan];
				}
			}
		}
		if ($fdpr[$chan] != -1) {
			if ($vflg > 1) {
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d select: Add channel %d proxy side\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan;
			}
			$sel->add($fdpr[$chan]);
			$cnt++;
		}
	}
	if ($cnt == 0) {
		# everything is closed
		if ($vflg) {
			@lt=localtime(time());
			printf "%04d-%02d-%02d %02d:%02d:%02d All descriptors closed\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0];
		}
		exit 0;
	}
	if ($vflg && $cnt == 1 && $fdsock != -1) {
		@lt=localtime(time());
		printf "%04d-%02d-%02d %02d:%02d:%02d Listening on port %d\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $listenport;
	}
	# wait for input
	@ready = $sel->can_read;
	foreach $fi (@ready) {
		# a fd is ready
		if ($vflg > 1) {
			@lt=localtime(time());
			printf "%04d-%02d-%02d %02d:%02d:%02d Input from socket/fd %d\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $fi;
		}
		# check listener
		if ($fi == $fdsock) {
			$fa=$fdsock->accept();
			if ($vflg) {
				$gpa=getpeername($fa);
				($gpt, $gia)=sockaddr_in($gpa);
				$gip=inet_ntoa($gia);
				if ($nflg == 0) {
					if ($ghn=gethostbyaddr($gia, AF_INET)) {
						$gip .= "($ghn)";
					}
				}
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d Incoming connection from %s", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $gip;
				if ($vflg > 1) {
					printf " remoteport=%d", $gpt;
				}
				printf "\n";
			}
			# if no looping
			if ($lflg == 0) {
				# close listener
				close($fdsock);
				$fdsock = -1;
			}
			# find free channel
			for ($chan=0; $chan < $chanmax; $chan++) {
				if ($chstate[$chan] == 0) {
					last;
				}
			}
			# hit upper limit
			if ($chan == $chanmax) {
				# extend limit
				$chanmax++;
			}
			$fdir[$chan]=$fa;
			$fdiw[$chan]=$fa;
			# make connection to proxy
			if ($vflg) {
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d channel %d connecting to proxy %s:%d\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $proxyhost, $proxyport;
			}
			unless ($fo = IO::Socket::INET->new(PeerAddr => $proxyhost, PeerPort => $proxyport, Proto => 'tcp')) {
				printf "$prog: Cannot connect to %s: %s", $proxyhp, $!."\n";
				close($fa);
				$fdir[$chan] = -1;
				$fdiw[$chan] = -1;
				$fdpr[$chan] = -1;
				$fdpw[$chan] = -1;
				$chstate[$chan]=0;
				last;
			}
			$fdpr[$chan]=$fo;
			$fdpw[$chan]=$fo;
			if ($vflg) {
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d channel %d created\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan;
			}
			# reset counter
			$cnti[$chan]=0;
			$cntp[$chan]=0;
			# send proxy request to remote connection
			printf $fo "CONNECT %s HTTP/1.0\r\n", $remotehp;
			if ($auth) {
				printf $fo "Proxy-authorization: basic %s\r\n", $auth;
			}
			printf $fo "\r\n";
			$chstate[$chan]=1;
			last;
		}

		# it is an file descriptor, find associated channel
		$rdwr=0;
		for ($chan=0; $chan < $chanmax; $chan++) {
			if ($fi == $fdir[$chan]) {
				$fo=$fdpw[$chan];
				$rdwr=1;
				$vside="input";
				last;
			}
			if ($fi == $fdpr[$chan]) {
				$fo=$fdiw[$chan];
				$rdwr=2;
				$vside="proxy";
				last;
			}
		}
		# read data
		$data="";
		$rstat=recv($fi, $data, $rdbufsize, 0);
		$lendata=length($data);
		if ($rdwr == 0) {
			# we did not find that descriptor in any channel
			if ($vflg > 1) {
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d unknown channel (fd=%d) read %d bytes\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $fi, $lendata;
			}
			# we have a design problem
#			exit 1;
			# just eat the data
			next;
		}
		if ($rdwr == 1) {
			$cnti[$chan] += $lendata;
		}
		if ($rdwr == 2) {
			$cntp[$chan] += $lendata;
		}
		if ($vflg > 1) {
			@lt=localtime(time());
			printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s side, read %d bytes\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside, $lendata;
		}
		if ($fo == -1) {
			# input trying to send to a closed write side
			if ($vflg > 1) {
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s read %d bytes, write side closed\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside, $lendata;
			}
			# just eat the data
			next;
		}
		# data to send
		if ($lendata > 0) {
			if ($vflg > 1) {
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s read %d bytes, state=%d\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside, $lendata, $chstate[$chan];
			}
			if ($chstate[$chan] != 4 && $rdwr == 2) {
				# eat web proxy initialization, searching for blank line
				for ($p=0; $p < $lendata; $p++) {
					$och=ord(substr($data, $p, 1));
					if ($vflg > 2) {
						@lt=localtime(time());
						printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s state=%d input[%d]=%d \"%s\"\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside, $chstate[$chan], $p, $och, substr($data, $p, 1);
					}
					if ($chstate[$chan] == 1) {
						if ($och == 10) {
							$chstate[$chan]=2;
						}
					} elsif ($chstate[$chan] == 2) {
						if ($och == 10) {
							$chstate[$chan]=4;
						} elsif ($och == 13) {
							$chstate[$chan]=3;
						} else {
							$chstate[$chan]=1;
						}
					} elsif ($chstate[$chan] == 3) {
						if ($och == 10) {
							$chstate[$chan]=4;
						} else {
							$chstate[$chan]=1;
						}
					}
					if ($chstate[$chan] == 4) {
						$lendata -= $p+1;
						$data=substr($data, $p);
						if ($vflg > 2) {
							@lt=localtime(time());
							printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s state=%d found sequence at %d\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside, $chstate[$chan], $p;
						}
						last;
					}
				}
				if ($chstate[$chan] != 4) {
					# did not find it
					next;
				}
				if ($vflg) {
					@lt=localtime(time());
					printf "%04d-%02d-%02d %02d:%02d:%02d channel %d webproxy startup ok\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan;
				}
				if ($lendata == 0) {
					# Found it, but ate all data
					next;
				}
			}
			# write the data
			if ($vflg > 1) {
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s writing %d bytes\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside, $lendata;
			}
			print $fo $data;
			next;
		}
		# no data means it was a close request
		if ($rdwr == 1) {
			# input side did a close
			if ($vflg) {
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s shutdown\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside;
			}
			shutdown($fdir[$chan], 0);
			$fdir[$chan] = -1;
			shutdown($fdpw[$chan], 1);
			$fdpw[$chan] = -1;
			if ($fdpr[$chan] == -1) {
				# proxy side closed already
				$chstate[$chan]=0;
				if ($vflg) {
					@lt=localtime(time());
					printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s finished (state=0)\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside;
				}
			}
		} else {
			# proxy side did a close
			if ($vflg) {
				@lt=localtime(time());
				printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s shutdown\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside;
			}
			shutdown($fdpr[$chan], 0);
			$fdpr[$chan] = -1;
			shutdown($fdiw[$chan], 1);
			$fdiw[$chan] = -1;
			if ($fdir[$chan] == -1) {
				# input side closed already
				$chstate[$chan]=0;
			}
		}
		if ($vflg && $chstate[$chan] == 0) {
			@lt=localtime(time());
			printf "%04d-%02d-%02d %02d:%02d:%02d channel %d %s finished. Sent proxy %d bytes, received %d bytes\n", $lt[5]+1900, $lt[4]+1, $lt[3], $lt[2], $lt[1], $lt[0], $chan, $vside, $cnti[$chan], $cntp[$chan];
		}
	}
}
